/* Much of this code was copied from 
 * http://gamedev.tutsplus.com/tutorials/implementation/speed-up-a-star-pathfinding-with-the-jump-point-search-algorithm/
 * which was incredibly useful. This project is designed for instructional purposes only.
 */

package com.mann.jump;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Set;
import java.util.Vector;

public class JumpPoint {
	
	private JumpPointGrid grid;
	private PriorityQueue<JumpPointCell> openSet;
	private HashSet<JumpPointCell> closedSet, jumpPoints, jumpPointsOnPath;
	private HashMap<JumpPointCell, JumpPointCell> cameFrom;
	private int expansionCount;
	private int jumpExpansionCount;
	
	public JumpPoint(JumpPointGrid grid) {
		this.grid = grid;
		closedSet = new HashSet<JumpPointCell>();
		openSet = new PriorityQueue<JumpPointCell>();
		cameFrom = new HashMap<JumpPointCell, JumpPointCell>();
		jumpPoints = new HashSet<JumpPointCell>();
		jumpPointsOnPath = new HashSet<JumpPointCell>();
		
		expansionCount = 0;
		jumpExpansionCount = 0;
	}
	
	private void initializeSearch() {
		grid.reinitializeGrid();
		
		closedSet.clear();
		openSet.clear();
		cameFrom.clear();
		jumpPoints.clear();
		jumpPointsOnPath.clear();
		
		expansionCount = 0;
		jumpExpansionCount = 0;
		
		grid.getStart().setG(0);
		grid.getStart().setF(grid.getHeuristic(grid.getStart()));
		openSet.add(grid.getStart());
	}
	
	/** The Jump Point Search algorithm works exactly like the A* search
	 * algorithm except that the get neighbors of A* is replaced with finding
	 * the nearest jump point for the current node.
	 * @return A path of jump points to take to reach the goal.
	 */
	public List<JumpPointCell> JPSearch() {
		initializeSearch();

		JumpPointCell current = null;
		while(!openSet.isEmpty()) {
			expansionCount++;
			current = openSet.poll();

			if(current.equals(grid.getGoal())) {
				return getPath();
			}

			closedSet.add(current);

			Vector<JumpPointCell> neighbors = identifySuccessors(current, grid.getStart(), grid.getGoal());
			for(JumpPointCell neighbor : neighbors) {
				double tentativeGScore = current.getG() + grid.getCost(current, neighbor);
				if(closedSet.contains(neighbor) && 
						tentativeGScore >= neighbor.getG()) {
					continue;
				}

				if(!openSet.contains(neighbor) || tentativeGScore < neighbor.getG()) {
					cameFrom.remove(neighbor);
					cameFrom.put(neighbor, current);
					neighbor.setG(tentativeGScore);
					neighbor.setF(neighbor.getG() + grid.getHeuristic(neighbor));
					openSet.remove(neighbor);
					openSet.add(neighbor);
				}
			}
		}
		return null;
	}
	
	public List<JumpPointCell> AStarSearch() {
		initializeSearch();

		JumpPointCell current = null;
		while(!openSet.isEmpty()) {
			expansionCount++;
			current = openSet.poll();

			if(current.equals(grid.getGoal())) {
				return getPath();
			}

			closedSet.add(current);

			List<JumpPointCell> neighbors = grid.getNonWallNeighbors(current);
			for(JumpPointCell neighbor : neighbors) {
				jumpExpansionCount++;
				double tentativeGScore = current.getG() + grid.getCost(current, neighbor);
				if(closedSet.contains(neighbor) && 
						tentativeGScore >= neighbor.getG()) {
					continue;
				}

				if(!openSet.contains(neighbor) || tentativeGScore < neighbor.getG()) {
					cameFrom.remove(neighbor);
					cameFrom.put(neighbor, current);
					neighbor.setG(tentativeGScore);
					neighbor.setF(neighbor.getG() + grid.getHeuristic(neighbor));
					openSet.remove(neighbor);
					openSet.add(neighbor);
				}
			}
		}
		return null;
	}
	
	public List<JumpPointCell> getPath() {
		LinkedList<JumpPointCell> path = new LinkedList<JumpPointCell>();
		
		/* Trace our way back through the parent map. */
		JumpPointCell current = grid.getGoal();
		jumpPointsOnPath.add(current);
		JumpPointCell next = cameFrom.get(current);
		while(next != null && !current.equals(grid.getStart())){
			int dX = next.getX() - current.getX();
			if(dX != 0) {
				dX /= Math.abs(dX);
			}
			int dY = next.getY() - current.getY();
			if(dY != 0) {
				dY /= Math.abs(dY);
			}
			while(!current.equals(next)) {
				path.add(current);
				int curX = current.getX(), curY = current.getY();
				int nextX = curX + dX, nextY = curY + dY;
				current = grid.get(nextX, nextY);
				if(current == null || current.isWall()) {
					return null;
				}
			}
			current = next;
			next = cameFrom.get(next);
			jumpPointsOnPath.add(current);
		}
		return path;
	}
	
	/** Returns the set of all jump points for the previously run search. */
	public Set<JumpPointCell> getJumpPoints() {
		return this.jumpPoints;
	}
	
	public void clearJumpPoints() {
		this.jumpPoints.clear();
		this.jumpPointsOnPath.clear();
	}
	
	/** Returns a set of all the jump points on the current path. */
	public Set<JumpPointCell> getJumpPointsOnPath() {
		return this.jumpPointsOnPath;
	}
	
	public int getCellExpansions() {
		return expansionCount;
	}
	
	public int getJumpExpansions() {
		return jumpExpansionCount;
	}
	
	private Vector<JumpPointCell> getPrunedNeighbors(JumpPointCell current, int dX, int dY) {
		Vector<JumpPointCell> neighbors = new Vector<JumpPointCell>(6);
		int nextX = current.getX() + dX;
		int nextY = current.getY() + dY;
		
		/* Initialize the list of neighbors to check. */
		JumpPointCell temp = null;
		
		/* The diagonal case has a lot of cells that we need to check, in fact
		 * there are 3 natural neighbors and possibly 2 forced neighbors that
		 * we need to look for.
		 */
		if(dX != 0 && dY != 0) {
			/* Add the natural neighbors to the vector. */
			temp = grid.get(current.getX(), current.getY()+dY);
			if(temp != null && !temp.isWall()) {
				neighbors.add(temp);
			}
			temp = grid.get(current.getX()+dX, current.getY());
			if(temp != null && !temp.isWall()) {
				neighbors.add(temp);
			}
			temp = grid.get(current.getX()+dX, current.getY()+dY);
			if(temp != null && !temp.isWall()) {
				neighbors.add(temp);
			}
			
			/* These are the forced neighbor checks. It is necessary in the 
			 * diagonal case to enforce that only one of the possible forced
			 * neighbors can exist (otherwise we cannot reach either neighbor
			 * using only the cardinal directions ie. N S E W).
			 */
			boolean xCheck = grid.isWall(current.getX(), current.getY()-dY) && 
					!grid.isWall(nextX, current.getY()-dY) &&
					grid.isValid(nextX, current.getY()-dY);
			boolean yCheck = grid.isWall(current.getX()-dX, current.getY()) && 
					!grid.isWall(current.getX()-dX, nextY) &&
					grid.isValid(current.getX()-dX, nextY);
			
			/*boolean onlyOne = (xCheck ^ yCheck);
			xCheck = xCheck && onlyOne;
			yCheck = yCheck && onlyOne;*/
			
			/* Add the forced neighbors to the vector. */
			if(xCheck) {
				neighbors.add(grid.get(nextX, current.getY()-dY));
			}
			if(yCheck) {
				neighbors.add(grid.get(current.getX()-dX, nextY));
			}
		} else if(dX != 0) {
			/* We know that nextY must be the current y value, but this is 
			 * here just as a precaution. */
			nextY = current.getY();
			
			/* Add all possible neighbors, forced and natural. */
			temp = grid.get(nextX, current.getY());
			boolean forwardCheck = temp != null && !temp.isWall();
			if(forwardCheck) {
				neighbors.add(temp);
			}
			
			/* Forced neighbor checks. */
			boolean upCheck = grid.isWall(current.getX(), nextY-1) && 
					!grid.isWall(nextX, nextY-1) &&
					grid.isValid(nextX, nextY-1);
			boolean downCheck = grid.isWall(current.getX(), nextY+1) && 
					!grid.isWall(nextX, nextY+1) &&
					grid.isValid(nextX, nextY+1);
			
			/*upCheck = upCheck && forwardCheck;
			downCheck = downCheck && forwardCheck;*/
			
			/* Add any forced neighbors. */
			if(upCheck) {
				neighbors.add(grid.get(nextX, nextY-1));
			}
			if(downCheck) {
				neighbors.add(grid.get(nextX, nextY+1));
			}

		} else if(dY != 0) {
			/* We know that nextX must be the current x value, but this is
			 * here just as a precaution.
			 */
			nextX = current.getX();
			
			/* We add the natural neighbor of the current cell. */
			temp = grid.get(current.getX(), nextY);
			boolean forwardCheck = temp != null && !temp.isWall();
			if(forwardCheck) {
				neighbors.add(temp);
			}
			
			/* Forced neighbor checks. */
			boolean leftCheck = grid.isWall(nextX-1, current.getY()) && 
					!grid.isWall(nextX-1, nextY) &&
					grid.isValid(nextX-1, nextY);
			boolean rightCheck = grid.isWall(nextX+1, current.getY()) && 
					!grid.isWall(nextX+1, nextY) &&
					grid.isValid(nextX+1, nextY);
			
			/*leftCheck = leftCheck && forwardCheck;
			rightCheck = rightCheck && forwardCheck;*/
			
			/* Add forced neighbors. */
			if(leftCheck) {
				neighbors.add(grid.get(nextX-1, nextY));
			}
			if(rightCheck) {
				neighbors.add(grid.get(nextX+1, nextY));
			}
		}
		
		jumpExpansionCount++;
		
		return neighbors;
	}
	
	private Vector<JumpPointCell> identifySuccessors(JumpPointCell current, JumpPointCell start, JumpPointCell end) {
		Vector<JumpPointCell> successors = new Vector<JumpPointCell>();
		int dX, dY;
		
		Vector<JumpPointCell> neighbors = new Vector<JumpPointCell>();
		JumpPointCell parent = cameFrom.get(current);
		if(parent == null) {
			for(JumpPointCell cell : grid.getNeighbors(current)) {
				if(!cell.isWall()) {
					neighbors.add(cell);
				}
			}
		} else {
			dX = current.getX() - parent.getX();
			if(dX != 0) {
				dX /= Math.abs(dX);
			}
			dY = current.getY() - parent.getY();
			if(dY != 0) {
				dY /= Math.abs(dY);
			}
			neighbors.addAll(getPrunedNeighbors(current, dX, dY));
		}
		
		for(JumpPointCell neighbor : neighbors) {
			/* These will determine the direction from our current cell to
			 * the neighboring cell. */
			dX = (neighbor.getX() - current.getX());
			dY = (neighbor.getY() - current.getY());
			
			/* We look for the jump points in the directions of our 
			 * neighbors. */
			JumpPointCell jumpPoint = jump(dX, dY, current, start, end);
			
			/* If we find a jump point add it to the list. */
			if(jumpPoint != null) {
				successors.add(jumpPoint);
			}
		}
		
		/* Add the jumpPoints to the set of all explore Jump Points! */
		jumpPoints.addAll(successors);
		
		/* Return the list of identified jump points. */
		return successors;
	}
	
	private JumpPointCell jump(int dX, int dY, JumpPointCell current, JumpPointCell start, JumpPointCell end) {
		/* Get the next cell that we are going to consider as a candidate
		 * jump point.
		 */
		jumpExpansionCount++;
		int nextX = current.getX() + dX;
		int nextY = current.getY() + dY;
		
		/* Check if there is a wall in the direction we are moving. */
		if(!grid.isValid(nextX, nextY) || grid.isWall(nextX, nextY)) {
			return null;
		}
		
		JumpPointCell next = grid.get(nextX, nextY);
		
		/* The direction we are moving in is actually the goal. */
		if(next.equals(end)) {
			return next;
		}
		
		if(dX != 0 && dY != 0) {
			/* We check for a forced diagonal neighbor. */
			boolean xCheck = grid.isWall(current.getX(), current.getY()-dY) && 
					!grid.isWall(nextX, current.getY()-dY) &&
					grid.isValid(nextX, current.getY()-dY);
			boolean yCheck = grid.isWall(current.getX()-dX, current.getY()) && 
					!grid.isWall(current.getX()-dX, nextY) &&
					grid.isValid(current.getX()-dX, nextY);
			
			/*boolean onlyOne = xCheck ^ yCheck;
			boolean forwardCheck = !(grid.isWall(current.getX(), nextY) && grid.isWall(nextX, current.getY()));*/
			
			if(xCheck || yCheck) {
				return next;
			}
			
			/* Special case for the diagonal, we need to check the vertical
			 * and horizontal directions before continuing diagonally. */
			if(jump(dX, 0, next, start, end) != null ||
			   jump(0, dY, next, start, end) != null) {
				return next;
			}
		} else {
			if(dX != 0) {
				/* Horizontal forced neighbor check should be true if either
				 * of the vertical neighbors of the node are walls. */
				boolean upCheck = grid.isWall(nextX, nextY-1) && 
						!grid.isWall(nextX+dX, nextY-1) &&
						grid.isValid(nextX+dX, nextY-1);
				boolean downCheck = grid.isWall(nextX, nextY+1) && 
						!grid.isWall(nextX+dX, nextY+1) &&
						grid.isValid(nextX+dX, nextY+1);
				
				/*JumpPointCell temp = grid.get(nextX, current.getY());
				boolean forwardCheck = temp != null && !temp.isWall();
				upCheck = upCheck && forwardCheck;
				downCheck = downCheck && forwardCheck;*/
				
				if(upCheck || downCheck){
					return next;
				}
			} else {
				/* We have found a jump point if we have a neighbor that is 
				 * forced due to obstacles. */
				boolean leftCheck = grid.isWall(nextX-1, nextY) && 
						!grid.isWall(nextX-1, nextY+dY) &&
						grid.isValid(nextX-1, nextY+dY);
				boolean rightCheck = grid.isWall(nextX+1, nextY) && 
						!grid.isWall(nextX+1, nextY+dY) &&
						grid.isValid(nextX+1, nextY+dY);
				
				/*JumpPointCell temp = grid.get(current.getX(), nextY);
				boolean forwardCheck = temp != null && !temp.isWall();
				leftCheck = forwardCheck && leftCheck;
				rightCheck = forwardCheck && rightCheck;*/
				
				if(leftCheck || rightCheck) {
					return next;
				}
			}
		}
		
		/* We take another step in the same direction and look for another
		 * jump point.
		 */
		return jump(dX, dY, next, start, end);
	}

}
