package com.mann.jump;

import java.util.LinkedList;
import java.util.List;

public class JumpPointGrid {
	
	private JumpPointCell[][] cellGrid;
	private JumpPointCell start, goal;
	
	public static final char WALL = '*', WEIGHT = '-';
	
	public JumpPointGrid(String[] contents) {
		this(contents.length, contents[0].length());
		for(int i = 0; i < cellGrid.length; i++) {
			for(int j = 0; j < cellGrid[0].length; j++) {
				if(contents[i].charAt(j) == WALL) {
					cellGrid[i][j].setWall(true);
				} else if(contents[i].charAt(j) == WEIGHT) {
					cellGrid[i][j].setDestWall(true);
				}
			}
		}
	}
	
	public JumpPointGrid(int length, int height) {
		this.cellGrid = new JumpPointCell[length][height];
		for(int i = 0; i < length; i++) {
			for(int j = 0; j < height; j++) {
				this.cellGrid[i][j] = new JumpPointCell(i, j);
			}
		}
	}
	
	/** This method will return the Manhattan estimated distance from the 
	 * cell to the goal.
	 * @param cell The cell for which you want the heuristic.
	 * @return The heuristic for the cell.
	 */
	public double getHeuristic(JumpPointCell cell) {
		double heuristic = Math.sqrt(Math.pow(goal.getX() - cell.getX(), 2) +
				Math.pow(goal.getY() - cell.getY(), 2));
		
		return heuristic;
	}
	
	/** This function assumes that the cells on either side of the edge
	 * are in direct sight of one another (no wall in the shortest path 
	 * between them).
	 * @param origin The start cell.
	 * @param dest The ending cell.
	 * @return The cost to take the edge from the start cell to the ending
	 * cell.
	 */
	public double getCost(JumpPointCell origin, JumpPointCell dest) {
		double dX = Math.pow(dest.getX() - origin.getX(), 2);
		double dY = Math.pow(dest.getY() - origin.getY(), 2);
		double cost = Math.sqrt(dX + dY);
		return cost;
	}
	
	public JumpPointCell get(int x, int y) {
		if(!isValid(x, y)) {
			return null;
		}
		return cellGrid[x][y];
	}
	
	public void setWall(int x, int y) {
		if(!isValid(x, y)) {
			return;
		}
		cellGrid[x][y].setWall(true);
	}
	
	public void removeWall(int x, int y) {
		if(!isValid(x, y)) {
			return;
		}
		cellGrid[x][y].setWall(false);
	}
	
	public boolean isValid(int x, int y) {
		return x < cellGrid.length && x >= 0 &&
				y < cellGrid[0].length && y >= 0;
	}
	
	public boolean isWall(int x, int y) {
		return isValid(x, y) && cellGrid[x][y].isWall();
	}
	
	public int getLength() {
		return cellGrid.length;
	}
	
	public int getHeight() {
		return cellGrid[0].length;
	}
	
	public void reinitializeGrid() {
		for(int i = 0; i < cellGrid.length; i++) {
			for(int j = 0; j < cellGrid[0].length; j++) {
				JumpPointCell current = cellGrid[i][j];
				current.setF(Double.POSITIVE_INFINITY);
				current.setG(Double.POSITIVE_INFINITY);
				if(current.isDestWall()) {
					current.setDestWall(true);
				}
			}
		}
	}
	
	public List<JumpPointCell> getNeighbors(JumpPointCell cell) {
		LinkedList<JumpPointCell> neighbors = new LinkedList<JumpPointCell>();
		
		int lowerX = Math.max(cell.getX() - 1, 0);
		int upperX = Math.min(cell.getX() + 2, cellGrid.length);
		int lowerY = Math.max(cell.getY() - 1, 0);
		int upperY = Math.min(cell.getY() + 2, cellGrid[0].length);
		
		for(int i = lowerX; i < upperX; i++) {
			for(int j = lowerY; j < upperY; j++) {
				neighbors.add(cellGrid[i][j]);
			}
		}
		neighbors.remove(cell);
		
		return neighbors;
	}
	
	public List<JumpPointCell> getNonWallNeighbors(JumpPointCell cell) {
		LinkedList<JumpPointCell> ret = new LinkedList<JumpPointCell>();
		for(JumpPointCell c : getNeighbors(cell)) {
			if(!c.isWall()){
				ret.add(c);
			}
		}
		
		/*JumpPointCell temp = get(cell.getX()+1, cell.getY());
		if(temp != null && !temp.isWall()) {
			ret.add(temp);
		}
		
		temp = get(cell.getX()-1, cell.getY());
		if(temp != null && !temp.isWall()) {
			ret.add(temp);
		}
		
		temp = get(cell.getX(), cell.getY()-1);
		if(temp != null && !temp.isWall()) {
			ret.add(temp);
		}
		
		temp = get(cell.getX(), cell.getY()+1);
		if(temp != null && !temp.isWall()) {
			ret.add(temp);
		}*/
		
		return ret;
	}
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		JumpPointCell current = null;
		
		for(int i = 0; i < cellGrid.length; i++) {
			for(int j = 0; j < cellGrid[0].length; j++) {
				current = cellGrid[i][j];
				double val = current.getF();
				if(val == Double.POSITIVE_INFINITY) {
					sb.append("\tinf\t");
				} else {
					sb.append("\t" + ((int)val) + "\t");
				}
			}
			sb.append(System.getProperty("line.separator"));
		}
		
		return sb.toString();
	}
	
	public JumpPointCell getStart() {
		return this.start;
	}
	
	public JumpPointCell getGoal() {
		return this.goal;
	}
	
	public void setStart(int row, int col) {
		if(!isValid(row, col)) {
			return;
		}
		this.start = cellGrid[row][col];
	}
	
	public void setGoal(int row, int col) {
		if(!isValid(row, col)) {
			return;
		}
		this.goal = cellGrid[row][col];
	}

}
