package com.mann.jump;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.List;
import java.util.Set;

import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;


public class JumpPointMain extends JApplet implements ActionListener {

	private static final long serialVersionUID = 2844160078566700618L;
	
	private BufferedImage image;
	private JPanel imagePanel;
	private JumpPointGrid grid;
	private JumpPoint solver;
	private List<JumpPointCell> curPath;
	private JButton solve, setCell;
	
	private static final int IMAGE_WIDTH = 801, IMAGE_HEIGHT = 801;
	private static final int CELL_WIDTH = 40, CELL_HEIGHT = 40;
	
	private static final String SET_START = "Start", SET_GOAL = "Goal", SET_WALL = "Wall", REMOVE_WALL = "RemoveWall";
	
	private static final String[] initialGrid = {
		"                **  ",
		"* ***     ***       ",
		"                *   ",
		"**    ***   *   *   ",
		"  *   *     *   *  *",
		"  ***       *   ****",
		"       *            ",
		"      ***       *** ",
		"                 *  ",
		"*****            * *",
		"                    ",
		"*     ***    **     ",
		"                    ",
		"   *    *           ",
		"   *    *   **     *",
		"  ***   **    *   **",
		"                    ",
		"** **           * * ",
		"     **          *  ",
		"          ***    *  "
	};
	
	public void init() {
		image = new BufferedImage(IMAGE_WIDTH, IMAGE_HEIGHT, BufferedImage.TYPE_INT_RGB);
		grid = new JumpPointGrid(initialGrid);
		solver = new JumpPoint(grid);
		curPath = null;
		
		imagePanel = new JPanel(){
			private static final long serialVersionUID = 1L;

			protected void paintComponent(Graphics g) {
				super.paintComponent(g);
				g.drawImage(image, 0, 0, null);
			}
		};
		add(imagePanel, BorderLayout.CENTER);
		
		JPanel buttonPanel = new JPanel();
		solve = new JButton("Start");
		solve.addActionListener(this);
		buttonPanel.add(solve, BorderLayout.EAST);
		setCell = new JButton("Set Cell");
		setCell.addActionListener(this);
		buttonPanel.add(setCell, BorderLayout.WEST);
		add(buttonPanel, BorderLayout.NORTH);
		
		setSize(IMAGE_WIDTH, IMAGE_HEIGHT + 36);
		setVisible(true);
		
		drawGrid();
	}
	
	/** Solves the current search problem defined in the grid object. */
	public void solve() {
		if(grid.getGoal() == null || grid.getStart() == null) {
			return;
		}
		reinitialize();
		curPath = solver.JPSearch();
	}
	
	/** Draws the grid onto the buffered image. */
	public void drawGrid() {
		/* We first white out the image, then we draw over it. */
		Graphics g = image.getGraphics();
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, IMAGE_WIDTH, IMAGE_HEIGHT);
		g.setColor(Color.BLACK);
		
		Set<JumpPointCell> jumpPoints = solver.getJumpPoints();
		if(jumpPoints != null && !jumpPoints.isEmpty()) {
			g.setColor(Color.PINK);
			for(JumpPointCell cell : jumpPoints) {
				g.fillRect(cell.getY()*CELL_HEIGHT, cell.getX()*CELL_WIDTH, CELL_WIDTH, CELL_HEIGHT);
			}
			g.setColor(Color.BLACK);
		}
		
		if(curPath != null && !curPath.isEmpty()) {
			g.setColor(Color.YELLOW);
			for(JumpPointCell cell : curPath) {
				g.fillRect(cell.getY()*CELL_HEIGHT, cell.getX()*CELL_WIDTH, CELL_WIDTH, CELL_HEIGHT);
			}
			g.setColor(Color.BLACK);
		}
		
		Set<JumpPointCell> jumpPointsOnPath = solver.getJumpPointsOnPath();
		if(jumpPointsOnPath != null && !jumpPointsOnPath.isEmpty()) {
			g.setColor(Color.GREEN);
			for(JumpPointCell cell : jumpPointsOnPath) {
				g.fillRect(cell.getY()*CELL_HEIGHT, cell.getX()*CELL_WIDTH, CELL_WIDTH, CELL_HEIGHT);
			}
			g.setColor(Color.BLACK);
		}
		
		if(grid.getStart() != null) {
			JumpPointCell current = grid.getStart();
			g.setColor(Color.RED);
			g.fillRect(current.getY()*CELL_HEIGHT, current.getX()*CELL_WIDTH, CELL_WIDTH, CELL_HEIGHT);
			g.setColor(Color.BLACK);
		}
		
		if(grid.getGoal() != null) {
			JumpPointCell current = grid.getGoal();
			g.setColor(Color.CYAN);
			g.fillRect(current.getY()*CELL_HEIGHT, current.getX()*CELL_WIDTH, CELL_WIDTH, CELL_HEIGHT);
			g.setColor(Color.BLACK);
		}
		
		for(int i = 0; i < grid.getLength(); i++) {
			for(int j = 0; j < grid.getHeight(); j++) {
				int xCoord = j*CELL_HEIGHT;
				int yCoord = i*CELL_WIDTH;
				if(grid.isWall(i, j)) {
					g.setColor(Color.LIGHT_GRAY);
					g.fillRect(xCoord, yCoord, CELL_WIDTH, CELL_HEIGHT);
				}
				g.setColor(Color.BLACK);
				g.drawRect(xCoord, yCoord, CELL_WIDTH, CELL_HEIGHT);
			}
		}
		this.repaint();
	}
	
	public void changeCell(int row, int column, String change) {
		if(change.equals(SET_WALL)) {
			grid.setWall(row, column);
		} else if(change.equals(SET_START)) {
			grid.setStart(row, column);
		} else if(change.equals(SET_GOAL)) {
			grid.setGoal(row, column);
		} else if(change.equals(REMOVE_WALL)) {
			grid.removeWall(row, column);
		}
		
		reinitialize();
		drawGrid();
	}
	
	public void reinitialize() {
		grid.reinitializeGrid();
		solver.clearJumpPoints();
		if(curPath != null) {
			curPath.clear();
		}
	}
	
	private class CustomDialog extends JDialog implements ActionListener, PropertyChangeListener {
		private static final long serialVersionUID = 1L;
		private JTextField rowField, columnField;
		private JComboBox<String> dropDown;

		private JOptionPane optionPane;

		private static final String enterString = "Enter";
		private static final String cancelString = "Cancel";

		/** Creates the reusable dialog. */
		public CustomDialog() {
			super();

			setTitle("Set Cell");

			rowField = new JTextField(5);
			columnField = new JTextField(5);
			String[] dropDownList = { SET_START, SET_GOAL, SET_WALL, REMOVE_WALL };

			//Create the combo box, and select Start.
			dropDown = new JComboBox<String>(dropDownList);
			dropDown.setSelectedIndex(0);

			//Create an array of the text and components to be displayed.
			String rowString = "Enter Row", colString = "Enter Column";
			Object[] array = {rowString, rowField, colString, columnField, dropDown};

			//Create an array specifying the number of dialog buttons
			//and their text.
			Object[] options = {enterString, cancelString};

			//Create the JOptionPane.
			optionPane = new JOptionPane(array,
					JOptionPane.QUESTION_MESSAGE,
					JOptionPane.YES_NO_OPTION,
					null,
					options,
					options[0]);

			//Make this dialog display it.
			setContentPane(optionPane);

			//Handle window closing correctly.
			setDefaultCloseOperation(DISPOSE_ON_CLOSE);

			//Ensure the text field always gets the first focus.
			addComponentListener(new ComponentAdapter() {
				public void componentShown(ComponentEvent ce) {
					rowField.requestFocusInWindow();
				}
			});

			//Register an event handler that puts the text into the option pane.
			rowField.addActionListener(this);
			columnField.addActionListener(this);

			//Register an event handler that reacts to option pane state changes.
			optionPane.addPropertyChangeListener(this);
			
			this.setSize(300, 200);
		}

		/** This method handles events for the text field. */
		public void actionPerformed(ActionEvent e) {
			optionPane.setValue(enterString);
		}

		/** This method reacts to state changes in the option pane. */
		public void propertyChange(PropertyChangeEvent e) {
			if (isVisible()) {
				Object value = optionPane.getValue();

				int row = -1, column = -1;
				if (enterString.equals(value)) {
					try {
						row = Integer.parseInt(rowField.getText());
						column = Integer.parseInt(columnField.getText());
					} catch(Exception e1) {
						e1.printStackTrace();
						clearAndHide();
						return;
					}
					clearAndHide();
					JumpPointMain.this.changeCell(row, column, (String)dropDown.getSelectedItem());
				} else {
					clearAndHide();
				}
			}
		}

		/** This method clears the dialog and hides it. */
		public void clearAndHide() {
			rowField.setText(null);
			columnField.setText(null);
			setVisible(false);
		}
	}
	
	@Override
	public void actionPerformed(ActionEvent event) {
		if(solve.equals(event.getSource())){
			solve();
			drawGrid();
		} else if(setCell.equals(event.getSource())) {
			new CustomDialog().setVisible(true);
		}
	}

}
